@@ -1,8 +1,10 @@
 import base64
 import io
 import json
-import licensespring
+
 import requests
+
+import licensespring
 from licensespring.api.authorization import (
     autorization_headers,
     date_header_value,
@@ -67,8 +69,8 @@ class APIClient:
 
     def request_additional_data(self, data, additional_data):
         for key, value in additional_data.items():
-            if key is not None:
-                data["key"] = value
+            if value is not None:
+                data[key] = value
         return data
 
     def request_hardware_data(
@@ -77,8 +79,6 @@ class APIClient:
         os_ver=None,
         hostname=None,
         ip=None,
-        is_vm=None,
-        vm_info=None,
         mac_address=None,
     ):
 
@@ -91,17 +91,21 @@ class APIClient:
             else self.hardware_id_provider.get_hostname()
         )
         data["ip"] = ip if ip is not None else self.hardware_id_provider.get_ip()
+        data["mac_address"] = (
+            mac_address
+            if mac_address is not None
+            else self.hardware_id_provider.get_mac_address()
+        )
+
+        return data
+
+    def request_vm_data(self, data, is_vm=None, vm_info=None):
         data["is_vm"] = (
             is_vm if is_vm is not None else self.hardware_id_provider.get_is_vm()
         )
         data["vm_info"] = (
             vm_info if vm_info is not None else self.hardware_id_provider.get_vm_info()
         )
-        data["mac_address"] = (
-            mac_address
-            if mac_address is not None
-            else self.hardware_id_provider.get_mac_address()
-        )
 
         return data
 
@@ -152,10 +156,9 @@ class APIClient:
             os_ver=os_ver,
             hostname=hostname,
             ip=ip,
-            is_vm=is_vm,
-            vm_info=vm_info,
             mac_address=mac_address,
         )
+        data = self.request_vm_data(data, is_vm=is_vm, vm_info=vm_info)
         data = self.request_version_data(data, app_ver=app_ver)
 
         response = self.send_request(
@@ -167,8 +170,8 @@ class APIClient:
 
         if self.verify_license_signature:
             self.signature_verifier.verify_license_signature(
-                hardware_id,
-                license_key,
+                hardware_id if hardware_id else self.hardware_id_provider.get_id(),
+                license_key if license_key else username,
                 response_json.get("validity_period"),
                 response_json.get("license_signature"),
             )
@@ -202,8 +205,8 @@ class APIClient:
 
         if self.verify_license_signature:
             self.signature_verifier.verify_license_signature(
-                hardware_id,
-                license_key,
+                hardware_id if hardware_id else self.hardware_id_provider.get_id(),
+                license_key if license_key else username,
                 response_json.get("validity_period"),
                 response_json.get("license_signature"),
             )
@@ -234,10 +237,9 @@ class APIClient:
             os_ver=os_ver,
             hostname=hostname,
             ip=ip,
-            is_vm=is_vm,
-            vm_info=vm_info,
             mac_address=mac_address,
         )
+        data = self.request_vm_data(data, is_vm=is_vm, vm_info=vm_info)
         data = self.request_version_data(data, app_ver=app_ver)
 
         data["api_key"] = self.api_key
@@ -249,6 +251,67 @@ class APIClient:
 
         return io.StringIO(base64.b64encode(json.dumps(data).encode()).decode())
 
+    def deactivate_offline(
+        self,
+        product,
+        hardware_id=None,
+        license_key=None,
+        username=None,
+        password=None,
+        app_ver=None,
+        os_ver=None,
+        hostname=None,
+        ip=None,
+        mac_address=None,
+        consumptions=None,
+        product_features=None,
+    ):
+        data = {"request": "deactivation"}
+        data = self.request_generic_data(
+            data, product, hardware_id, license_key, username, password
+        )
+        data = self.request_hardware_data(
+            data=data,
+            os_ver=os_ver,
+            hostname=hostname,
+            ip=ip,
+            mac_address=mac_address,
+        )
+        data = self.request_version_data(data, app_ver=app_ver)
+
+        data["api_key"] = self.api_key
+        data["date"] = date_header_value()
+        data["signature"] = offline_signature(
+            self.api_key, self.shared_key, data["hardware_id"], license_key, username
+        )
+        data["request_id"] = self.hardware_id_provider.get_request_id()
+        data = self.request_additional_data(
+            data=data,
+            additional_data={
+                "consumptions": consumptions,
+                "product_features": product_features,
+            },
+        )
+
+        return io.StringIO(base64.b64encode(json.dumps(data).encode()).decode())
+
+    def activate_offline_load(self, f, hardware_id=None):
+        value = f.getvalue()
+        decoded_value = base64.b64decode(value).decode()
+        response_json = json.loads(decoded_value)
+
+        if self.verify_license_signature:
+            self.signature_verifier.verify_license_signature(
+                hardware_id if hardware_id else self.hardware_id_provider.get_id(),
+                response_json.get("license_key")
+                if response_json.get("license_key")
+                else response_json.get("username"),
+                response_json.get("validity_period"),
+                response_json.get("license_signature"),
+            )
+
+        return response_json
+
     def add_consumption(
         self,
         product,
@@ -276,7 +339,7 @@ class APIClient:
         response = self.send_request(
             method="post",
             endpoint="/add_consumption",
-            json=data,
+            data=data,
         )
 
         return response.json()
@@ -306,7 +369,7 @@ class APIClient:
         response = self.send_request(
             method="post",
             endpoint="/add_feature_consumption",
-            json=data,
+            data=data,
         )
 
         return response.json()
@@ -349,7 +412,7 @@ class APIClient:
         response = self.send_request(
             method="get",
             endpoint="/trial_key",
-            json=data,
+            data=data,
         )
 
         return response.json()
@@ -363,7 +426,7 @@ class APIClient:
         response = self.send_request(
             method="get",
             endpoint="/product_details",
-            json=data,
+            data=data,
         )
 
         return response.json()
@@ -385,7 +448,7 @@ class APIClient:
         response = self.send_request(
             method="post",
             endpoint="/track_device_variables",
-            json=data,
+            jsdataon=data,
         )
 
         return True if response.status_code == 200 else False
@@ -405,7 +468,7 @@ class APIClient:
         response = self.send_request(
             method="post",
             endpoint="/floating/release",
-            json=data,
+            data=data,
         )
 
         return True if response.status_code == 200 else False
@@ -428,7 +491,7 @@ class APIClient:
         response = self.send_request(
             method="post",
             endpoint="/floating/borrow",
-            json=data,
+            data=data,
         )
 
         return response.json()
@@ -443,7 +506,7 @@ class APIClient:
         response = self.send_request(
             method="post",
             endpoint="/change_password",
-            json=data,
+            data=data,
         )
 
         return True if response.status_code == 200 else False
@@ -470,7 +533,7 @@ class APIClient:
         response = self.send_request(
             method="get",
             endpoint="/versions",
-            json=data,
+            data=data,
         )
 
         return response.json()
